# Copyright (c) 2020 Vlsarro
# Copyright (c) 2013 Calin Crisan
# This file is part of motionEye.
#
# motionEye is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from motioneye.utils import GetCamerasResponse
from motioneye.utils.http import RtmpUrl

__all__ = ('test_rtmp_url',)


def test_rtmp_url(data: dict) -> GetCamerasResponse:
    url_obj = RtmpUrl.from_dict(data)

    # Since RTMP is a binary TCP stream its a little more work to do a proper test
    # For now lets just check if a TCP socket is open on the target IP:PORT
    # TODO: Actually do the TCP SYN/ACK check...

    cameras = [{'id': 'tcp', 'name': 'RTMP/TCP Camera'}]
    return GetCamerasResponse(cameras, None)



# Copyright (c) 2020 Vlsarro
# Copyright (c) 2013 Calin Crisan
# This file is part of motionEye.
#
# motionEye is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import socket

from motioneye.handlers.base import BaseHandler
from motioneye.motionctl import find_motion
from motioneye.update import get_os_version

__all__ = ('VersionHandler',)


class VersionHandler(BaseHandler):
    def get(self):
        motion_info = find_motion()
        os_version = get_os_version()

        self.render(
            'version.html',
            os_version=' '.join(os_version),
            motion_version=motion_info[1] if motion_info else '',
            hostname=socket.gethostname(),
        )

    post = get



# Copyright (c) 2020 Vlsarro
# Copyright (c) 2013 Calin Crisan
# This file is part of motionEye.
#
# motionEye is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import base64
import hashlib
import logging
import os
import re
import subprocess
import sys
import time
import typing
import urllib.error
import urllib.parse
import urllib.request
from collections import namedtuple
from dataclasses import dataclass

from PIL import Image, ImageDraw
from tornado.concurrent import Future
from tornado.ioloop import IOLoop

from motioneye import settings

_SIGNATURE_REGEX = re.compile(r'[^a-zA-Z0-9/?_.=&{}\[\]":, -]')
_SPECIAL_COOKIE_NAMES = {'expires', 'domain', 'path', 'secure', 'httponly'}

MASK_WIDTH = 32

DEV_NULL = open('/dev/null', 'w')

COMMON_RESOLUTIONS = [
    (320, 200),
    (320, 240),
    (640, 480),
    (800, 480),
    (800, 600),
    (1024, 576),
    (1024, 600),
    (1024, 768),
    (1280, 720),
    (1280, 768),
    (1280, 800),
    (1280, 960),
    (1280, 1024),
    (1440, 900),
    (1440, 960),
    (1440, 1024),
    (1600, 1200),
    (1920, 1080),
    (3840, 2160),
]

GetCamerasResponse = namedtuple('GetCamerasResponse', ('cameras', 'error'))
GetConfigResponse = namedtuple('GetConfigResponse', ('remote_ui_config', 'error'))
GetMotionDetectionResult = namedtuple('GetMotionDetectionResult', ('enabled', 'error'))


@dataclass
class GetCurrentPictureResponse:
    motion_detected: bool = False
    capture_fps: typing.Any = None
    monitor_info: typing.Any = None
    picture: typing.Any = None
    error: typing.Any = None


@dataclass
class ListMediaResponse:
    media_list: list = None
    error: str = None


@dataclass
class CommonExternalResponse:
    result: typing.Any = None
    error: typing.Any = None


def cast_future(obj: typing.Awaitable[typing.Any]) -> Future:
    return typing.cast(Future, obj)


def spawn_callback_timeout_wrapper(
    callback: typing.Callable, *args: typing.Any, **kwargs: typing.Any
) -> None:
    IOLoop.current().spawn_callback(callback, *args, **kwargs)


def pretty_size(size):
    if size < 1024:  # less than 1kB
        size, unit = size, 'B'

    elif size < 1024 * 1024:  # less than 1MB
        size, unit = size / 1024.0, 'kB'

    elif size < 1024 * 1024 * 1024:  # less than 1GB
        size, unit = size / 1024.0 / 1024, 'MB'

    else:  # greater than or equal to 1GB
        size, unit = size / 1024.0 / 1024 / 1024, 'GB'

    return f'{size:.1f} {unit}'


def pretty_http_error(response):
    if response.code == 401 or response.error == 'Authentication Error':
        return 'authentication failed'

    if not response.error:
        return 'ok'

    msg = str(response.error)
    if msg.startswith('HTTP '):
        msg = msg.split(':', 1)[-1].strip()

    if msg.startswith('[Errno '):
        msg = msg.split(']', 1)[-1].strip()

    if 'timeout' in msg.lower() or 'timed out' in msg.lower():
        msg = 'request timed out'

    return msg


def make_str(s):
    if isinstance(s, str):
        return s

    try:
        return str(s)

    except:
        try:
            return str(s, encoding='utf8').encode('utf8')

        except:
            return str(s).encode('utf8')


def split_semicolon(s):
    parts = s.split(';')
    merged_parts = []
    for p in parts:
        if merged_parts and merged_parts[-1][-1] == '\\':
            merged_parts[-1] = merged_parts[-1][:-1] + ';' + p

        else:
            merged_parts.append(p)

    if not merged_parts:
        return []

    return [p.strip() for p in merged_parts]


def get_disk_usage(path):
    logging.debug(f'getting disk usage for path {path}...')

    try:
        result = os.statvfs(path)

    except OSError as e:
        logging.error(f'failed to execute statvfs: {str(e)}')

        return None

    block_size = result.f_frsize
    free_blocks = result.f_bfree
    total_blocks = result.f_blocks

    free_size = free_blocks * block_size
    total_size = total_blocks * block_size
    used_size = total_size - free_size

    return used_size, total_size


def is_local_motion_camera(config):
    """Tells if a camera is managed by the local motion instance."""
    return bool(
        config.get('videodevice')
        or config.get('video_device')
        or config.get('netcam_url')
        or config.get('mmalcam_name')
    )


def is_remote_camera(config):
    """Tells if a camera is managed by a remote motionEye server."""
    return config.get('@proto') == 'motioneye'


def is_v4l2_camera(config):
    """Tells if a camera is a v4l2 device managed by the local motion instance."""
    return bool(config.get('videodevice'))


def is_mmal_camera(config):
    """Tells if a camera is mmal device managed by the local motion instance."""
    return bool(config.get('mmalcam_name'))


def is_net_camera(config):
    """Tells if a camera is a network camera managed by the local motion instance."""
    return bool(config.get('netcam_url'))


def is_simple_mjpeg_camera(config):
    """Tells if a camera is a simple MJPEG camera not managed by any motion instance."""
    return bool(config.get('@proto') == 'mjpeg')


def compute_signature(method, path, body: bytes, key):
    parts = list(urllib.parse.urlsplit(path))
    query = [
        q
        for q in urllib.parse.parse_qsl(parts[3], keep_blank_values=True)
        if (q[0] != '_signature')
    ]
    query.sort(key=lambda q: q[0])
    # "safe" characters here are set to match the encodeURIComponent JavaScript counterpart
    query = [(n, urllib.parse.quote(v, safe="!'()*~")) for (n, v) in query]
    query = '&'.join([(q[0] + '=' + q[1]) for q in query])
    parts[0] = parts[1] = ''
    parts[3] = query
    path = urllib.parse.urlunsplit(parts)
    path = _SIGNATURE_REGEX.sub('-', path)
    key = _SIGNATURE_REGEX.sub('-', key)

    try:
        body_str = body.decode('utf-8')
    except:
        body_str = None

    if body_str and body_str.startswith('---'):
        body_str = None  # file attachment

    body_str = body_str and _SIGNATURE_REGEX.sub('-', body_str)

    return (
        hashlib.sha1(
            ('{}:{}:{}:{}'.format(method, path, body_str or '', key)).encode('utf-8')
        )
        .hexdigest()
        .lower()
    )


def parse_cookies(cookies_headers):
    parsed = {}

    for cookie in cookies_headers:
        cookie = cookie.split(';')
        for c in cookie:
            (name, value) = c.split('=', 1)
            name = name.strip()
            value = value.strip()

            if name.lower() in _SPECIAL_COOKIE_NAMES:
                continue

            parsed[name] = value

    return parsed


def build_basic_header(username, password):
    return 'Basic %s' % base64.b64encode(f'{username}:{password}'.encode()).decode()


def parse_basic_header(header):
    parts = header.split(' ', 1)
    if len(parts) < 2:
        return None

    if parts[0].lower() != 'basic':
        return None

    encoded = parts[1]

    try:
        decoded = base64.decodebytes(encoded)

    except:
        return None

    parts = decoded.split(':', 1)
    if len(parts) < 2:
        return None

    return {'username': parts[0], 'password': parts[1]}


def build_digest_header(method, url, username, password, state):
    realm = state['realm']
    nonce = state['nonce']
    last_nonce = state.get('last_nonce', '')
    nonce_count = state.get('nonce_count', 0)
    qop = state.get('qop')
    algorithm = state.get('algorithm')
    opaque = state.get('opaque')

    if algorithm is None:
        _algorithm = 'MD5'

    else:
        _algorithm = algorithm.upper()

    if _algorithm == 'MD5' or _algorithm == 'MD5-SESS':

        def md5_utf8(x):
            if isinstance(x, str):
                x = x.encode('utf-8')
            return hashlib.md5(x).hexdigest()

        hash_utf8 = md5_utf8

    else:  # _algorithm == 'SHA'

        def sha_utf8(x):
            if isinstance(x, str):
                x = x.encode('utf-8')
            return hashlib.sha1(x).hexdigest()

        hash_utf8 = sha_utf8

    KD = lambda s, d: hash_utf8(f"{s}:{d}")

    if hash_utf8 is None:
        return None

    entdig = None
    p_parsed = urllib.parse.urlparse(url)
    path = p_parsed.path
    if p_parsed.query:
        path += '?' + p_parsed.query

    A1 = f'{username}:{realm}:{password}'
    A2 = f'{method}:{path}'

    HA1 = hash_utf8(A1)
    HA2 = hash_utf8(A2)

    if nonce == last_nonce:
        nonce_count += 1

    else:
        nonce_count = 1

    ncvalue = '%08x' % nonce_count
    s = str(nonce_count).encode('utf-8')
    s += nonce.encode('utf-8')
    s += time.ctime().encode('utf-8')
    s += os.urandom(8)

    cnonce = hashlib.sha1(s).hexdigest()[:16]
    if _algorithm == 'MD5-SESS':
        HA1 = hash_utf8(f'{HA1}:{nonce}:{cnonce}')

    if qop is None:
        respdig = KD(HA1, f"{nonce}:{HA2}")

    elif qop == 'auth' or 'auth' in qop.split(','):
        noncebit = "{}:{}:{}:{}:{}".format(nonce, ncvalue, cnonce, 'auth', HA2)
        respdig = KD(HA1, noncebit)

    else:
        return None

    last_nonce = nonce

    base = 'username="%s", realm="%s", nonce="%s", uri="%s", ' 'response="%s"' % (
        username,
        realm,
        nonce,
        path,
        respdig,
    )
    if opaque:
        base += ', opaque="%s"' % opaque
    if algorithm:
        base += ', algorithm="%s"' % algorithm
    if entdig:
        base += ', digest="%s"' % entdig
    if qop:
        base += f', qop=auth, nc={ncvalue}, cnonce="{cnonce}"'

    state['last_nonce'] = last_nonce
    state['nonce_count'] = nonce_count

    return 'Digest %s' % base


def urlopen(*args, **kwargs):
    if sys.version_info >= (2, 7, 9) and not settings.VALIDATE_CERTS:
        # ssl certs are not verified by default
        # in versions prior to 2.7.9

        import ssl

        ctx = ssl.create_default_context()
        ctx.check_hostname = False
        ctx.verify_mode = ssl.CERT_NONE

        kwargs.setdefault('context', ctx)

    return urllib.request.urlopen(*args, **kwargs)


def build_editable_mask_file(
    camera_id, mask_class, mask_lines, capture_width=None, capture_height=None
):
    if not mask_lines:
        return ''

    width = mask_lines[0]
    height = mask_lines[1]
    mask_lines = mask_lines[2:]

    logging.debug(
        'building editable %s mask for camera with id %s (%sx%s)'
        % (mask_class, camera_id, width, height)
    )

    # horizontal rectangles
    nx = MASK_WIDTH  # number of rectangles
    if width % nx:
        nx -= 1
        rx = width % nx  # remainder

    else:
        rx = 0

    rw = width / nx  # rectangle width

    # vertical rectangles
    ny = mask_height = height * MASK_WIDTH // width  # number of rectangles
    if height % ny:
        ny -= 1
        ry = height % ny  # remainder

    else:
        ry = 0

    # if mask not present, generate an empty mask
    if not mask_lines:
        mask_lines = [0] * mask_height

    # scale the mask vertically in case the aspect ratio has changed
    # since the last time the mask has been generated
    if ny == len(mask_lines):
        line_index_func = lambda y: y

    else:
        line_index_func = lambda y: (len(mask_lines) - 1) * y / ny

    rh = height / ny  # rectangle height

    # draw the actual mask image content
    im = Image.new('L', (width, height), 255)  # all white
    dr = ImageDraw.Draw(im)

    for y in range(ny):
        line = mask_lines[int(line_index_func(y))]
        for x in range(nx):
            if line & (1 << (MASK_WIDTH - 1 - x)):
                dr.rectangle(
                    (x * rw, y * rh, (x + 1) * rw - 1, (y + 1) * rh - 1), fill=0
                )

        if rx and line & 1:
            dr.rectangle((nx * rw, y * rh, nx * rw + rx - 1, (y + 1) * rh - 1), fill=0)

    if ry:
        line = mask_lines[int(line_index_func(ny))]
        for x in range(nx):
            if line & (1 << (MASK_WIDTH - 1 - x)):
                dr.rectangle(
                    (x * rw, ny * rh, (x + 1) * rw - 1, ny * rh + ry - 1), fill=0
                )

        if rx and line & 1:
            dr.rectangle((nx * rw, ny * rh, nx * rw + rx - 1, ny * rh + ry - 1), fill=0)

    #    file_name = os.path.join(settings.CONF_PATH, 'mask_%s.pgm' % camera_id)
    file_name = build_mask_file_name(camera_id, mask_class)

    # resize the image if necessary
    if capture_width and capture_height and im.size != (capture_width, capture_height):
        logging.debug(
            'editable mask needs resizing from %sx%s to %sx%s'
            % (im.size[0], im.size[1], capture_width, capture_height)
        )

        im = im.resize((capture_width, capture_height))

    im.save(file_name, 'ppm')

    return file_name


def build_mask_file_name(camera_id, mask_class):
    file_name = (
        'mask_%s.pgm' % (camera_id)
        if mask_class == 'motion'
        else f'mask_{camera_id}_{mask_class}.pgm'
    )
    full_path = os.path.join(settings.CONF_PATH, file_name)

    return full_path


def parse_editable_mask_file(
    camera_id, mask_class, capture_width=None, capture_height=None
):
    # capture_width and capture_height arguments represent the current size
    # of the camera image, as it might be different from that of the associated mask;
    # they can be null (e.g. netcams)

    file_name = build_mask_file_name(camera_id, mask_class)

    logging.debug(
        f'parsing editable mask {mask_class} for camera with id {camera_id}: {file_name}'
    )

    # read the image file
    try:
        im = Image.open(file_name)

    except Exception as e:
        logging.error(f'failed to read mask file {file_name}: {e}')

        # empty mask
        return [0] * (MASK_WIDTH * 10)

    if capture_width and capture_height:
        # resize the image if necessary
        if im.size != (capture_width, capture_height):
            logging.debug(
                'editable mask needs resizing from %sx%s to %sx%s'
                % (im.size[0], im.size[1], capture_width, capture_height)
            )

            im = im.resize((capture_width, capture_height))

        width, height = capture_width, capture_height

    else:
        logging.debug(f'using mask size from file: {im.size[0]}x{im.size[1]}')

        width, height = im.size

    pixels = list(im.getdata())

    # horizontal rectangles
    nx = MASK_WIDTH  # number of rectangles
    if width % nx:
        nx -= 1
        rx = width % nx  # remainder

    else:
        rx = 0

    rw = width // nx  # rectangle width

    # vertical rectangles
    ny = height * MASK_WIDTH // width  # number of rectangles
    if height % ny:
        ny -= 1
        ry = height % ny  # remainder

    else:
        ry = 0

    rh = height // ny  # rectangle height

    # parse the image contents and build the mask lines
    mask_lines = [width, height]
    for y in range(ny):
        bits = []
        for x in range(nx):
            px = int((x + 0.5) * rw)
            py = int((y + 0.5) * rh)
            pixel = pixels[py * width + px]
            bits.append(not bool(pixel))

        if rx:
            px = int(nx * rw + rx / 2)
            py = int((y + 0.5) * rh)
            pixel = pixels[py * width + px]
            bits.append(not bool(pixel))

        # build the binary packed mask line
        line = 0
        for i, bit in enumerate(bits):
            if bit:
                line |= 1 << (MASK_WIDTH - 1 - i)

        mask_lines.append(line)

    if ry:
        bits = []
        for x in range(nx):
            px = int((x + 0.5) * rw)
            py = int(ny * rh + ry / 2)
            pixel = pixels[py * width + px]
            bits.append(not bool(pixel))

        if rx:
            px = int(nx * rw + rx / 2)
            py = int(ny * rh + ry / 2)
            pixel = pixels[py * width + px]
            bits.append(not bool(pixel))

        # build the binary packed mask line
        line = 0
        for i, bit in enumerate(bits):
            if bit:
                line |= 1 << (MASK_WIDTH - 1 - i)

        mask_lines.append(line)

    return mask_lines


def call_subprocess(
    args,
    stdin=None,
    input=None,
    stdout=subprocess.PIPE,
    stderr=DEV_NULL,
    capture_output=False,
    shell=False,
    cwd=None,
    timeout=None,
    check=True,
    encoding='utf-8',
    errors=None,
    text=None,
    env=None,
) -> str:
    """subprocess.run wrapper to return output as a decoded string"""
    return subprocess.run(
        args,
        stdin=stdin,
        input=input,
        stdout=stdout,
        stderr=stderr,
        capture_output=capture_output,
        shell=shell,
        cwd=cwd,
        timeout=timeout,
        check=check,
        encoding=encoding,
        errors=errors,
        text=text,
        env=env,
    ).stdout.strip()



# Copyright (c) 2013 Calin Crisan
# This file is part of motionEye.
#
# motionEye is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import atexit
import datetime
import logging
import multiprocessing
import os
import re
import signal
import sys
import time

from tornado.ioloop import IOLoop
from tornado.web import Application

from motioneye import settings, template
from motioneye.controls import smbctl, v4l2ctl
from motioneye.handlers.action import ActionHandler
from motioneye.handlers.base import ManifestHandler, NotFoundHandler
from motioneye.handlers.config import ConfigHandler
from motioneye.handlers.log import LogHandler
from motioneye.handlers.login import LoginHandler
from motioneye.handlers.main import MainHandler
from motioneye.handlers.movie import MovieHandler
from motioneye.handlers.movie_playback import MovieDownloadHandler, MoviePlaybackHandler
from motioneye.handlers.picture import PictureHandler
from motioneye.handlers.power import PowerHandler
from motioneye.handlers.prefs import PrefsHandler
from motioneye.handlers.relay_event import RelayEventHandler
from motioneye.handlers.update import UpdateHandler
from motioneye.handlers.version import VersionHandler

_PID_FILE = 'motioneye.pid'
_CURRENT_PICTURE_REGEX = re.compile(r'^/picture/\d+/current')


class Daemon:
    def __init__(self, pid_file, run_callback=None):
        self.pid_file = pid_file
        self.run_callback = run_callback

    def daemonize(self):
        # first fork
        try:
            if os.fork() > 0:  # parent
                sys.exit(0)

        except OSError as e:
            sys.stderr.write('fork() failed: %s\n' % e.strerror)
            sys.exit(-1)

        # separate from parent
        os.setsid()
        os.umask(0)

        # second fork
        try:
            if os.fork() > 0:  # parent
                sys.exit(0)

        except OSError as e:
            sys.stderr.write('fork() failed: %s\n' % e.strerror)
            sys.exit(-1)

            # redirect standard file descriptors
        sys.stdout.flush()
        sys.stderr.flush()
        si = open('/dev/null')
        so = open('/dev/null', 'a+')
        se = open('/dev/null', 'a+')
        os.dup2(si.fileno(), sys.stdin.fileno())
        os.dup2(so.fileno(), sys.stdout.fileno())
        os.dup2(se.fileno(), sys.stderr.fileno())

        # pid file
        atexit.register(self.del_pid)
        with open(self.pid_file, 'w') as f:
            f.write('%s\n' % os.getpid())

    def del_pid(self):
        try:
            os.remove(self.pid_file)

        except:
            sys.stderr.write('failed to remove pid file.\n')

    def running(self):
        try:
            with open(self.pid_file) as f:
                pid = int(f.read().strip())

        except:
            return None

        try:
            os.kill(pid, 0)
            return pid

        except:
            return None

    def start(self):
        if self.running():
            sys.stderr.write('server is already running\n')
            sys.exit(-1)

        self.daemonize()
        sys.stdout.write('server started\n')
        self.run_callback()

    def stop(self):
        pid = self.running()
        if not pid:
            sys.stderr.write('server is not running\n')
            sys.exit(-1)

        try:
            os.kill(pid, signal.SIGTERM)

        except Exception as e:
            sys.stderr.write('failed to terminate server: %s\n' % e)

        for i in range(50):  # @UnusedVariable
            try:
                os.kill(pid, 0)
                time.sleep(0.1)

            except OSError as e:
                if str(e).count('No such process'):
                    self.del_pid()
                    sys.stdout.write('server stopped\n')
                    break

                else:
                    sys.stderr.write('failed to terminate server: %s\n' % e)
                    sys.exit(-11)

        else:
            sys.stderr.write('server failed to stop, killing it\n')
            try:
                os.kill(pid, signal.SIGKILL)

            except:
                sys.stderr.write('failed to kill...\n')


def _log_request(handler):
    log_method = None

    if handler.get_status() < 400:
        if not _CURRENT_PICTURE_REGEX.match(handler.request.uri):
            log_method = logging.debug

    elif handler.get_status() < 500:
        log_method = logging.warning

    else:
        log_method = logging.error

    if log_method:
        request_time = 1000.0 * handler.request.request_time()
        log_method(
            "%d %s %.2fms",
            handler.get_status(),
            handler._request_summary(),
            request_time,
        )


handler_mapping = [
    (r'^/$', MainHandler),
    (r'^/manifest.json$', ManifestHandler),
    (r'^/config/main/(?P<op>set|get)/?$', ConfigHandler),
    (
        r'^/config/(?P<camera_id>\d+)/(?P<op>get|set|rem|test|authorize)/?$',
        ConfigHandler,
    ),
    (r'^/config/(?P<op>add|list|backup|restore)/?$', ConfigHandler),
    (r'^/picture/(?P<camera_id>\d+)/(?P<op>current|list|frame)/?$', PictureHandler),
    (
        r'^/picture/(?P<camera_id>\d+)/(?P<op>download|preview|delete)/(?P<filename>.+?)/?$',
        PictureHandler,
    ),
    (
        r'^/picture/(?P<camera_id>\d+)/(?P<op>zipped|timelapse|delete_all)/(?P<group>.*?)/?$',
        PictureHandler,
    ),
    (r'^/movie/(?P<camera_id>\d+)/(?P<op>list)/?$', MovieHandler),
    (
        r'^/movie/(?P<camera_id>\d+)/(?P<op>preview|delete)/(?P<filename>.+?)/?$',
        MovieHandler,
    ),
    (r'^/movie/(?P<camera_id>\d+)/(?P<op>delete_all)/(?P<group>.*?)/?$', MovieHandler),
    (
        r'^/movie/(?P<camera_id>\d+)/playback/(?P<filename>.+?)/?$',
        MoviePlaybackHandler,
        {'path': r''},
    ),
    (
        r'^/movie/(?P<camera_id>\d+)/download/(?P<filename>.+?)/?$',
        MovieDownloadHandler,
        {'path': r''},
    ),
    (r'^/action/(?P<camera_id>\d+)/(?P<action>\w+)/?$', ActionHandler),
    (r'^/prefs/(?P<key>\w+)?/?$', PrefsHandler),
    (r'^/_relay_event/?$', RelayEventHandler),
    (r'^/log/(?P<name>\w+)/?$', LogHandler),
    (r'^/update/?$', UpdateHandler),
    (r'^/power/(?P<op>shutdown|reboot)/?$', PowerHandler),
    (r'^/version/?$', VersionHandler),
    (r'^/login/?$', LoginHandler),
    (r'^.*$', NotFoundHandler),
]


def configure_signals():
    def bye_handler(signal, frame):
        logging.info(_('interrompa signalo ricevita, fermanta …'))

        # shut down the IO loop if it has been started
        io_loop = IOLoop.current()
        io_loop.stop()

    def child_handler(signal, frame):
        # this is required for the multiprocessing mechanism to work
        multiprocessing.active_children()

    signal.signal(signal.SIGINT, bye_handler)
    signal.signal(signal.SIGTERM, bye_handler)
    signal.signal(signal.SIGCHLD, child_handler)


def test_requirements():
    if not os.access(settings.CONF_PATH, os.W_OK):
        logging.fatal(
            'config directory "%s" does not exist or is not writable'
            % settings.CONF_PATH
        )
        sys.exit(-1)

    if not os.access(settings.RUN_PATH, os.W_OK):
        logging.fatal(
            'pid directory "%s" does not exist or is not writable' % settings.RUN_PATH
        )
        sys.exit(-1)

    if not os.access(settings.LOG_PATH, os.W_OK):
        logging.fatal(
            'log directory "%s" does not exist or is not writable' % settings.LOG_PATH
        )
        sys.exit(-1)

    if not os.access(settings.MEDIA_PATH, os.W_OK):
        logging.fatal(
            'media directory "%s" does not exist or is not writable'
            % settings.MEDIA_PATH
        )
        sys.exit(-1)

    if os.geteuid() != 0:
        if settings.SMB_SHARES:
            logging.fatal(_('smb-akcioj postulas radikajn privilegiojn'))
            sys.exit(-1)

    try:
        import tornado  # @UnusedImport

    except ImportError:
        logging.fatal(_('bonvolu instali tornado version 3.1 aŭ pli'))
        sys.exit(-1)

    try:
        import jinja2  # @UnusedImport

    except ImportError:
        logging.fatal(_('bonvolu instali jinja2'))
        sys.exit(-1)

    try:
        import PIL.Image  # @UnusedImport

    except ImportError:
        logging.fatal(_('bonvolu instali pillow aŭ PIL'))
        sys.exit(-1)

    try:
        import pycurl  # @UnusedImport

    except ImportError:
        logging.fatal(_('bonvolu instali pycurl'))
        sys.exit(-1)

    from motioneye import motionctl

    has_motion = motionctl.find_motion()[0] is not None

    from motioneye import mediafiles

    has_ffmpeg = mediafiles.find_ffmpeg() is not None

    has_v4lutils = v4l2ctl.find_v4l2_ctl() is not None

    if settings.SMB_SHARES and smbctl.find_mount_cifs() is None:
        logging.fatal('please install cifs-utils')
        sys.exit(-1)

    if not has_motion:
        logging.info('motion not installed')

    if not has_ffmpeg:
        if has_motion:
            logging.warning('you have motion installed, but no ffmpeg')

        else:
            logging.info('ffmpeg not installed')

    if not has_v4lutils:
        if has_motion:
            logging.warning('you have motion installed, but no v4l-utils')

        else:
            logging.info('v4l-utils not installed')


def make_media_folders():
    from motioneye import config

    config.get_main()  # just to have main config already loaded

    camera_ids = config.get_camera_ids()
    for camera_id in camera_ids:
        camera_config = config.get_camera(camera_id)
        if 'target_dir' in camera_config:
            if not os.path.exists(camera_config['target_dir']):
                try:
                    os.makedirs(camera_config['target_dir'])

                except Exception as e:
                    logging.error(
                        'failed to create root media folder "{}" for camera with id {}: {}'.format(
                            camera_config['target_dir'], camera_id, e
                        )
                    )


def start_motion():
    from motioneye import config, motionctl

    io_loop = IOLoop.current()

    # add a motion running checker
    def checker():
        if (
            not motionctl.running()
            and motionctl.started()
            and config.get_enabled_local_motion_cameras()
        ):
            try:
                logging.error('motion not running, starting it')
                motionctl.start()

            except Exception as e:
                logging.error(f'failed to start motion: {str(e)}', exc_info=True)

        io_loop.add_timeout(
            datetime.timedelta(seconds=settings.MOTION_CHECK_INTERVAL), checker
        )

    try:
        motionctl.start()

    except Exception as e:
        logging.error(str(e), exc_info=True)

    io_loop.add_timeout(
        datetime.timedelta(seconds=settings.MOTION_CHECK_INTERVAL), checker
    )


def parse_options(parser, args):
    parser.add_argument(
        '-b',
        help='start the server in background (daemonize)',
        action='store_true',
        dest='background',
        default=False,
    )

    return parser.parse_args(args)


def make_app(debug: bool = False) -> Application:
    return Application(
        handler_mapping,
        debug=debug,
        log_function=_log_request,
        static_path=settings.STATIC_PATH,
        static_url_prefix='/static/',
    )


def run():
    import motioneye
    from motioneye import cleanup, mjpgclient, motionctl, tasks, wsswitch
    from motioneye.controls import smbctl

    configure_signals()
    logging.info(_('saluton! ĉi tio estas motionEye-servilo ') + motioneye.VERSION)

    test_requirements()
    make_media_folders()

    if settings.SMB_SHARES:
        stop, start = smbctl.update_mounts()  # @UnusedVariable
        if start:
            start_motion()

    else:
        start_motion()

    if settings.CLEANUP_INTERVAL:
        cleanup.start()
        logging.info(_('purigado komenciĝis'))
    wsswitch.start()
    logging.info(_('wsswitch komenciĝis'))

    tasks.start()
    logging.info(_('taskoj komenciĝis'))

    if settings.MJPG_CLIENT_TIMEOUT:
        mjpgclient.start()
        logging.info(_('mjpg klienta rubo-kolektanto komenciĝis'))

    if settings.SMB_SHARES:
        smbctl.start()
        logging.info('smb mounts started')

    template.add_context('static_path', 'static/')
    template.add_context('lingvo', settings.lingvo)

    application = Application(
        handler_mapping,
        debug=False,
        log_function=_log_request,
        static_path=settings.STATIC_PATH,
        static_url_prefix='/static/',
    )

    application.listen(settings.PORT, settings.LISTEN)
    logging.info(_('servilo komenciĝis'))
    io_loop = IOLoop.current()
    # we need to reset the loop's PID to fix PID checks when running in daemon mode
    io_loop._pid = os.getpid()
    io_loop.start()

    logging.info(_('servilo haltis'))
    tasks.stop()
    logging.info(_('taskoj haltis'))

    if cleanup.running():
        cleanup.stop()
        logging.info('cleanup stopped')

    if motionctl.running():
        motionctl.stop()
        logging.info(_('motion haltis'))
    if settings.SMB_SHARES:
        smbctl.stop()
        logging.info('smb mounts stopped')

    logging.info(_('adiaŭ!'))


def main(parser, args, command):
    from motioneye import meyectl

    options = parse_options(parser, args)

    meyectl.configure_logging('motioneye', options.background or options.log_to_file)
    meyectl.configure_tornado()

    if command == 'start':
        if options.background:
            daemon = Daemon(
                pid_file=os.path.join(settings.RUN_PATH, _PID_FILE), run_callback=run
            )
            daemon.start()

        else:
            run()

    elif command == 'stop':
        daemon = Daemon(pid_file=os.path.join(settings.RUN_PATH, _PID_FILE))
        daemon.stop()



# Copyright (c) 2020 Vlsarro
# Copyright (c) 2013 Calin Crisan
# This file is part of motionEye.
#
# motionEye is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
from functools import cmp_to_key

from motioneye.handlers.base import BaseHandler
from motioneye.update import (
    compare_versions,
    get_all_versions,
    get_os_version,
    perform_update,
)

__all__ = ('UpdateHandler',)


class UpdateHandler(BaseHandler):
    @BaseHandler.auth(admin=True)
    def get(self):
        logging.debug('listing versions')

        versions = get_all_versions()
        current_version = get_os_version()[
            1
        ]  # os version is returned as (name, version) tuple
        recent_versions = [
            v for v in versions if compare_versions(v, current_version) > 0
        ]
        recent_versions.sort(key=cmp_to_key(compare_versions))
        update_version = recent_versions[-1] if recent_versions else None

        self.finish_json(
            {'update_version': update_version, 'current_version': current_version}
        )

    @BaseHandler.auth(admin=True)
    def post(self):
        version = self.get_argument('version')

        logging.debug(f'performing update to version {version}')

        result = perform_update(version)

        self.finish_json(result)



# Copyright (c) 2020 Vlsarro
# Copyright (c) 2013 Calin Crisan
# This file is part of motionEye.
#
# motionEye is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from motioneye import config, mediafiles, motionctl, tasks, uploadservices, utils
from motioneye.handlers.base import BaseHandler

__all__ = ('RelayEventHandler',)


class RelayEventHandler(BaseHandler):
    @BaseHandler.auth(admin=True)
    def post(self):
        event = self.get_argument('event')
        motion_camera_id = int(self.get_argument('motion_camera_id'))

        camera_id = motionctl.motion_camera_id_to_camera_id(motion_camera_id)
        if camera_id is None:
            logging.debug(
                'ignoring event for unknown motion camera id %s' % motion_camera_id
            )
            return self.finish_json()

        else:
            logging.debug(
                'received relayed event {event} for motion camera id {id} (camera id {cid})'.format(
                    event=event, id=motion_camera_id, cid=camera_id
                )
            )

        camera_config = config.get_camera(camera_id)
        if not utils.is_local_motion_camera(camera_config):
            logging.warning(
                'ignoring event for non-local camera with id %s' % camera_id
            )
            return self.finish_json()

        if event == 'start':
            if not camera_config['@motion_detection']:
                logging.debug(
                    'ignoring start event for camera with id %s and motion detection disabled'
                    % camera_id
                )
                return self.finish_json()

            motionctl.set_motion_detected(camera_id, True)

        elif event == 'stop':
            motionctl.set_motion_detected(camera_id, False)

        elif event == 'movie_end':
            filename = self.get_argument('filename')

            # generate preview (thumbnail)
            tasks.add(
                5,
                mediafiles.make_movie_preview,
                tag='make_movie_preview(%s)' % filename,
                camera_config=camera_config,
                full_path=filename,
            )

            # upload to external service
            if camera_config['@upload_enabled'] and camera_config['@upload_movie']:
                self.upload_media_file(filename, camera_id, camera_config)

        elif event == 'picture_save':
            filename = self.get_argument('filename')

            # upload to external service
            if camera_config['@upload_enabled'] and camera_config['@upload_picture']:
                self.upload_media_file(filename, camera_id, camera_config)

        else:
            logging.warning('unknown event %s' % event)

        self.finish_json()

    def upload_media_file(self, filename, camera_id, camera_config):
        service_name = camera_config['@upload_service']

        tasks.add(
            5,
            uploadservices.upload_media_file,
            tag='upload_media_file(%s)' % filename,
            camera_id=camera_id,
            service_name=service_name,
            camera_name=camera_config['camera_name'],
            target_dir=camera_config['@upload_subfolders']
            and camera_config['target_dir'],
            filename=filename,
        )



# Copyright (c) 2020 Vlsarro
# Copyright (c) 2013 Calin Crisan
# This file is part of motionEye.
#
# motionEye is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime

from tornado.ioloop import IOLoop

from motioneye.controls.powerctl import PowerControl
from motioneye.handlers.base import BaseHandler

__all__ = ('PowerHandler',)


class PowerHandler(BaseHandler):
    @BaseHandler.auth(admin=True)
    def post(self, op):
        if op == 'shutdown':
            self.shut_down()

        elif op == 'reboot':
            self.reboot()

    def shut_down(self):
        io_loop = IOLoop.current()
        io_loop.add_timeout(datetime.timedelta(seconds=2), PowerControl.shut_down)

    def reboot(self):
        io_loop = IOLoop.current()
        io_loop.add_timeout(datetime.timedelta(seconds=2), PowerControl.reboot)



# Copyright (c) 2013 Calin Crisan
# This file is part of motionEye.
#
# motionEye is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import ftplib
import io
import json
import logging
import mimetypes
import os
import os.path
import time
from base64 import b64encode
from urllib.error import HTTPError
from urllib.parse import quote, urlencode
from urllib.request import Request

import boto3
import pycurl

from motioneye import settings, utils

_STATE_FILE_NAME = 'uploadservices.json'
_services = None


class UploadService:
    MAX_FILE_SIZE = 1024 * 1024 * 1024  # 1GB

    NAME = 'base'

    def __init__(self, camera_id, **kwargs):
        self.camera_id = camera_id

    def __str__(self):
        return self.NAME

    @classmethod
    def get_authorize_url(cls):
        return '/'

    def test_access(self):
        return True

    def upload_file(self, target_dir, filename, camera_name):
        ctime = os.path.getctime(filename)

        if target_dir:
            target_dir = os.path.realpath(target_dir)
            rel_filename = os.path.realpath(filename)
            rel_filename = rel_filename[len(target_dir) :]

            while rel_filename.startswith('/'):
                rel_filename = rel_filename[1:]

            self.debug(f'uploading file "{target_dir}/{rel_filename}" to {self}')

        else:
            rel_filename = os.path.basename(filename)

            self.debug(f'uploading file "{filename}" to {self}')

        try:
            st = os.stat(filename)

        except Exception as e:
            msg = f'failed to open file "{filename}": {e}'
            self.error(msg)
            raise Exception(msg)

        if st.st_size > self.MAX_FILE_SIZE:
            msg = 'file "{}" is too large ({}MB/{}MB)'.format(
                filename,
                st.st_size / 1024 / 1024,
                self.MAX_FILE_SIZE / 1024 / 1024,
            )

            self.error(msg)
            raise Exception(msg)

        try:
            f = open(filename, 'rb')

        except Exception as e:
            msg = f'failed to open file "{filename}": {e}'
            self.error(msg)
            raise Exception(msg)

        data = f.read()
        self.debug(f'size of "{filename}" is {len(data) / 1024.0 / 1024:.3f}MB')

        mime_type = mimetypes.guess_type(filename)[0] or 'image/jpeg'
        self.debug(f'mime type of "{filename}" is "{mime_type}"')

        self.upload_data(rel_filename, mime_type, data, ctime, camera_name)

        self.debug(f'file "{filename}" successfully uploaded')

    def upload_data(self, filename, mime_type, data, ctime, camera_name):
        pass

    def dump(self):
        return {}

    def load(self, data):
        pass

    def save(self):
        services = _load()
        camera_services = services.setdefault(self.camera_id, {})
        camera_services[self.NAME] = self

        _save(services)

    def log(self, level, message, **kwargs):
        message = self.NAME + ': ' + message

        logging.log(level, message, **kwargs)

    def debug(self, message, **kwargs):
        self.log(logging.DEBUG, message, **kwargs)

    def info(self, message, **kwargs):
        self.log(logging.INFO, message, **kwargs)

    def error(self, message, **kwargs):
        self.log(logging.ERROR, message, **kwargs)

    def clean_cloud(self, cloud_dir, local_folders):
        pass

    @staticmethod
    def get_service_classes():
        return {c.NAME: c for c in UploadService.__subclasses__()}


class GoogleBase:
    AUTH_URL = 'https://accounts.google.com/o/oauth2/auth'
    TOKEN_URL = 'https://accounts.google.com/o/oauth2/token'

    CLIENT_ID = (
        '349038943026-m16svdadjrqc0c449u4qv71v1m1niu5o.apps.googleusercontent.com'
    )
    CLIENT_NOT_SO_SECRET = 'jjqbWmICpA0GvbhsJB3okX7s'

    def _init(self):
        self._location = None
        self._authorization_key = None
        self._credentials = None
        self._folder_ids = {}
        self._folder_id_times = {}

    @classmethod
    def _get_authorize_url(cls):
        query = {
            'scope': cls.SCOPE,
            'redirect_uri': 'urn:ietf:wg:oauth:2.0:oob',
            'response_type': 'code',
            'client_id': cls.CLIENT_ID,
            'access_type': 'offline',
        }

        return cls.AUTH_URL + '?' + urlencode(query)

    def _test_access(self):
        try:
            self._folder_ids = {}
            self._get_folder_id()
            return True

        except Exception as e:
            return str(e)

    def _dump(self):
        return {
            'location': self._location,
            'credentials': self._credentials,
            'authorization_key': self._authorization_key,
        }

    def _load(self, data):
        if data.get('location'):
            self._location = data['location']
            self._folder_ids = {}
        if data.get('authorization_key'):
            self._authorization_key = data['authorization_key']
            self._credentials = None
        if data.get('credentials'):
            self._credentials = data['credentials']

    def _request(self, url, body=None, headers=None, retry_auth=True, method=None):
        if not self._credentials:
            if not self._authorization_key:
                msg = 'missing authorization key'
                self.error(msg)
                raise Exception(msg)

            self.debug('requesting credentials')
            try:
                self._credentials = self._request_credentials(self._authorization_key)
                self.save()

            except Exception as e:
                self.error(f'failed to obtain credentials: {e}')
                raise

        headers = headers or {}
        headers['Authorization'] = f'Bearer {self._credentials["access_token"]}'

        self.debug(f'requesting {url}')
        request = Request(url, data=body, headers=headers)
        if method:
            request.get_method = lambda: method
        try:
            response = utils.urlopen(request)

        except HTTPError as e:
            if (
                e.code == 401 and retry_auth
            ):  # unauthorized, access token may have expired
                try:
                    self.debug('credentials have probably expired, refreshing them')
                    self._credentials = self._refresh_credentials(
                        self._credentials['refresh_token']
                    )
                    self.save()

                    # retry the request with refreshed credentials
                    return self._request(url, body, headers, retry_auth=False)

                except Exception:
                    self.error('refreshing credentials failed')
                    raise

            else:
                try:
                    e = json.load(e)
                    msg = e['error']['message']

                except Exception:
                    msg = str(e)

                self.error(f'request failed: {msg}')
                raise Exception(msg)

        except Exception as e:
            self.error(f'request failed: {e}')
            raise

        return response.read()

    def _request_json(self, url, body=None, headers=None, retry_auth=True, method=None):
        response = self._request(url, body, headers, retry_auth, method)
        try:
            response = json.loads(response)
        except Exception:
            self.error("response doesn't seem to be a valid json")
            raise

        return response

    def _request_credentials(self, authorization_key):
        headers = {'Content-Type': 'application/x-www-form-urlencoded'}

        body = {
            'code': authorization_key,
            'redirect_uri': 'urn:ietf:wg:oauth:2.0:oob',
            'client_id': self.CLIENT_ID,
            'client_secret': self.CLIENT_NOT_SO_SECRET,
            'scope': self.SCOPE,
            'grant_type': 'authorization_code',
        }
        body = urlencode(body).encode()

        request = Request(self.TOKEN_URL, data=body, headers=headers)

        try:
            response = utils.urlopen(request)

        except HTTPError as e:
            error = json.load(e)
            raise Exception(
                error.get('error_description') or error.get('error') or str(e)
            )

        data = json.load(response)

        return {
            'access_token': data['access_token'],
            'refresh_token': data['refresh_token'],
        }

    def _refresh_credentials(self, refresh_token):
        headers = {'Content-Type': 'application/x-www-form-urlencoded'}

        body = {
            'refresh_token': refresh_token,
            'client_id': self.CLIENT_ID,
            'client_secret': self.CLIENT_NOT_SO_SECRET,
            'grant_type': 'refresh_token',
        }
        body = urlencode(body).encode()

        request = Request(self.TOKEN_URL, data=body, headers=headers)

        try:
            response = utils.urlopen(request)

        except HTTPError as e:
            error = json.load(e)
            raise Exception(
                error.get('error_description') or error.get('error') or str(e)
            )

        data = json.load(response)

        return {
            'access_token': data['access_token'],
            'refresh_token': data.get('refresh_token', refresh_token),
        }


class GoogleDrive(UploadService, GoogleBase):
    NAME = 'gdrive'

    SCOPE = 'https://www.googleapis.com/auth/drive.file'
    CHILDREN_URL = (
        'https://www.googleapis.com/drive/v2/files/%(parent_id)s/children?q=%(query)s'
    )
    CHILDREN_QUERY = (
        "'%(parent_id)s' in parents and title = '%(child_name)s' and trashed = false"
    )
    UPLOAD_URL = 'https://www.googleapis.com/upload/drive/v2/files?uploadType=multipart'
    CREATE_FOLDER_URL = 'https://www.googleapis.com/drive/v2/files'

    BOUNDARY = 'motioneye_multipart_boundary'

    FOLDER_ID_LIFE_TIME = 300  # 5 minutes

    def __init__(self, camera_id):
        self._init()

        UploadService.__init__(self, camera_id)

    @classmethod
    def get_authorize_url(cls):
        return cls._get_authorize_url()

    def test_access(self):
        return self._test_access()

    def upload_data(self, filename, mime_type, data, ctime, camera_name):
        path = os.path.dirname(filename)
        filename = os.path.basename(filename)

        metadata = {'title': filename, 'parents': [{'id': self._get_folder_id(path)}]}

        body = [
            '--' + self.BOUNDARY,
            'Content-Type: application/json; charset=UTF-8',
            '',
            json.dumps(metadata),
            '',
            '--' + self.BOUNDARY,
            f'Content-Type: {mime_type}',
            '',
            '',
        ]

        body = '\r\n'.join(body).encode()
        body += data
        body += f'\r\n--{self.BOUNDARY}--'.encode()

        headers = {
            'Content-Type': f'multipart/related; boundary="{self.BOUNDARY}"',
            'Content-Length': len(body),
        }

        self._request(self.UPLOAD_URL, body, headers)

    def dump(self):
        return self._dump()

    def load(self, data):
        self._load(data)

    def _get_folder_id(self, path=''):
        now = time.time()

        folder_id = self._folder_ids.get(path)
        folder_id_time = self._folder_id_times.get(path, 0)

        location = self._location
        if not location.endswith('/'):
            location += '/'

        location += path

        if not folder_id or (now - folder_id_time > self.FOLDER_ID_LIFE_TIME):
            self.debug(f'finding folder id for location "{location}"')
            folder_id = self._get_folder_id_by_path(location)

            self._folder_ids[path] = folder_id
            self._folder_id_times[path] = now

        return folder_id

    def _get_folder_id_by_path(self, path):
        if path and path != '/':
            path = [p.strip() for p in path.split('/') if p.strip()]
            parent_id = 'root'
            for name in path:
                parent_id = self._get_folder_id_by_name(parent_id, name)

            return parent_id

        else:  # root folder
            return self._get_folder_id_by_name(None, 'root')

    def _get_folder_id_by_name(self, parent_id, child_name, create=True):
        if parent_id:
            query = self.CHILDREN_QUERY % {
                'parent_id': parent_id,
                'child_name': child_name,
            }
            query = quote(query)

        else:
            query = ''

        parent_id = parent_id or 'root'
        # when requesting the id of the root folder, we perform a dummy request,
        # event though we already know the id (which is "root"), to test the request

        url = self.CHILDREN_URL % {'parent_id': parent_id, 'query': query}
        response = self._request(url)
        try:
            response = json.loads(response)

        except Exception:
            self.error("response doesn't seem to be a valid json")
            raise

        if parent_id == 'root' and child_name == 'root':
            return 'root'

        items = response.get('items')
        if not items:
            if create:
                self.debug(
                    f'folder with name "{child_name}" does not exist, creating it'
                )
                self._create_folder(parent_id, child_name)
                return self._get_folder_id_by_name(parent_id, child_name, create=False)

            else:
                msg = f'folder with name "{child_name}" does not exist'
                self.error(msg)
                raise Exception(msg)

        return items[0]['id']

    def _create_folder(self, parent_id, child_name):
        metadata = {
            'title': child_name,
            'parents': [{'id': parent_id}],
            'mimeType': 'application/vnd.google-apps.folder',
        }

        body = json.dumps(metadata).encode()

        headers = {'Content-Type': 'application/json; charset=UTF-8'}

        self._request(self.CREATE_FOLDER_URL, body, headers)

    def clean_cloud(self, cloud_dir, local_folders):
        # remove old cloud folder that does not exist in local.
        # assumes 'cloud_dir' is a direct child of the 'root'.

        removed_count = 0
        folder_id = self._get_folder_id_by_name('root', cloud_dir, False)
        children = self._get_children(folder_id)
        self.info(
            f'found {len(local_folders)}/{len(children)} folder(s) in local/cloud'
        )
        self.debug(f'local {local_folders}')
        for child in children:
            id = child['id']
            name = self._get_file_title(id)
            self.debug(f"cloud '{name}'")
            to_delete = not exist_in_local(name, local_folders)
            if to_delete and self._delete_file(id):
                removed_count += 1
                self.info(f"deleted a cloud folder '{name}'")

        self.info(f'deleted {removed_count} cloud folder(s)')
        return removed_count

    def _get_children(self, file_id):
        url = f'{self.CREATE_FOLDER_URL}/{file_id}/children'
        response = self._request(url)

        try:
            response = json.loads(response)

        except Exception:
            self.error("response doesn't seem to be a valid json")
            raise

        return response['items']

    def _delete_file(self, file_id):
        url = f'{self.CREATE_FOLDER_URL}/{file_id}'
        response = self._request(url, None, None, True, 'DELETE')
        succeeded = response == ""
        return succeeded

    def _get_file_metadata(self, file_id):
        url = f'{self.CREATE_FOLDER_URL}/{file_id}'
        response = self._request(url)

        try:
            response = json.loads(response)

        except Exception:
            self.error("response doesn't seem to be a valid json")
            raise

        return response

    def _get_file_title(self, file_id):
        return self._get_file_metadata(file_id)['title']


class GooglePhoto(UploadService, GoogleBase):
    NAME = 'gphoto'

    SCOPE = 'https://www.googleapis.com/auth/photoslibrary'
    GOOGLE_PHOTO_API = 'https://photoslibrary.googleapis.com/v1/'

    def __init__(self, camera_id):
        self._init()

        UploadService.__init__(self, camera_id)

    @classmethod
    def get_authorize_url(cls):
        return cls._get_authorize_url()

    def test_access(self):
        return self._test_access()

    def upload_data(self, filename, mime_type, data, ctime, camera_name):
        path = os.path.dirname(filename)
        filename = os.path.basename(filename)
        dayinfo = datetime.datetime.fromtimestamp(ctime).strftime('%Y-%m-%d')
        uploadname = dayinfo + '-' + filename

        body = data

        headers = {
            'Content-Type': 'application/octet-stream',
            'X-Goog-Upload-File-Name': uploadname,
            'X-Goog-Upload-Protocol': 'raw',
        }

        uploadToken = self._request(self.GOOGLE_PHOTO_API + 'uploads', body, headers)
        response = self._create_media(uploadToken, camera_name)
        self.debug(f'response {response["mediaItem"]}')

    def dump(self):
        return self._dump()

    def load(self, data):
        self._load(data)

    def _get_folder_id(self, path=''):
        location = self._location

        folder_id = self._folder_ids.get(location)

        self.debug(f'_get_folder_id({path}, {location}, {folder_id})')

        if not folder_id:
            self.debug(f'finding album with title "{location}"')
            folder_id = self._get_folder_id_by_name(location)

            self._folder_ids[location] = folder_id

        return folder_id

    def _get_folder_id_by_name(self, name, create=True):
        try:
            albums = self._get_albums()
            albumsWithName = self._filter_albums(albums, name)

            if albumsWithName:
                count = len(albumsWithName)
                if count > 0:
                    albumId = albumsWithName[0].get('id')
                    self.debug(
                        f'found {count} existing album(s) "{name}" taking first id "{albumId}"'
                    )
                    return albumId

            # create album
            response = self._create_folder(None, name)
            albumId = response.get('id')
            self.info(f'Album "{name}" was created successfully with id "{albumId}"')
            return albumId

        except Exception as e:
            self.error(f"_get_folder_id_by_name() failed: {e}")
            raise

    def _create_folder(self, parent_id, child_name):
        metadata = {'album': {'title': child_name}}

        body = json.dumps(metadata)

        headers = {'Content-Type': 'application/json'}

        response = self._request_json(self.GOOGLE_PHOTO_API + 'albums', body, headers)
        return response

    def _create_media(self, uploadToken, camera_name):
        description = 'captured by motionEye camera' + (
            f' "{camera_name}"' if camera_name else ''
        )

        metadata = {
            'albumId': self._get_folder_id(),
            'newMediaItems': [
                {
                    'description': description,
                    'simpleMediaItem': {'uploadToken': uploadToken},
                }
            ],
        }

        body = json.dumps(metadata)

        headers = {'Content-Type': 'application/json'}

        response = self._request_json(
            self.GOOGLE_PHOTO_API + 'mediaItems:batchCreate', body, headers
        )
        return response.get('newMediaItemResults')[0]

    def _get_albums(self):
        response = self._request_json(self.GOOGLE_PHOTO_API + 'albums')

        albums = response.get('albums')
        self.debug(f'got {len(albums)} album(s)')
        return albums

    def _filter_albums(self, albums, title):
        return [a for a in albums if a.get('title') == title]


class Dropbox(UploadService):
    NAME = 'dropbox'

    AUTH_URL = 'https://www.dropbox.com/oauth2/authorize'
    TOKEN_URL = 'https://api.dropboxapi.com/oauth2/token'

    CLIENT_ID = 'dropbox_client_id_placeholder'
    CLIENT_NOT_SO_SECRET = 'dropbox_client_secret_placeholder'

    LIST_FOLDER_URL = 'https://api.dropboxapi.com/2/files/list_folder'
    UPLOAD_URL = 'https://content.dropboxapi.com/2/files/upload'

    def __init__(self, camera_id):
        self._location = None
        self._authorization_key = None
        self._credentials = None

        UploadService.__init__(self, camera_id)

    @classmethod
    def get_authorize_url(cls):
        query = {
            'response_type': 'code',
            'client_id': cls.CLIENT_ID,
            'token_access_type': 'offline',
        }

        return cls.AUTH_URL + '?' + urlencode(query)

    def test_access(self):
        body = {
            'path': self._clean_location(),
            'recursive': False,
            'include_media_info': False,
            'include_deleted': False,
        }

        body = json.dumps(body)
        headers = {'Content-Type': 'application/json'}

        try:
            self._request(self.LIST_FOLDER_URL, body, headers)
            return True

        except Exception as e:
            msg = str(e)

            # remove trailing punctuation
            while msg and not msg[-1].isalnum():
                msg = msg[:-1]

            return msg

    def upload_data(self, filename, mime_type, data, ctime, camera_name):
        metadata = {
            'path': os.path.join(self._clean_location(), filename),
            'mode': 'add',
            'autorename': True,
            'mute': False,
        }

        headers = {
            'Content-Type': 'application/octet-stream',
            'Dropbox-API-Arg': json.dumps(metadata),
        }

        self._request(self.UPLOAD_URL, data, headers)

    def dump(self):
        return {
            'location': self._location,
            'credentials': self._credentials,
            'authorization_key': self._authorization_key,
        }

    def load(self, data):
        if data.get('location'):
            self._location = data['location']
        if data.get('authorization_key'):
            self._authorization_key = data['authorization_key']
            self._credentials = None
        if data.get('credentials'):
            self._credentials = data['credentials']

    def _clean_location(self):
        location = self._location
        if location == '/':
            return ''

        if not location.startswith('/'):
            location = '/' + location

        return location

    def _request(self, url, body=None, headers=None, retry_auth=True):
        if not self._credentials:
            if not self._authorization_key:
                msg = 'missing authorization key'
                self.error(msg)
                raise Exception(msg)

            self.debug('requesting credentials')
            try:
                self._credentials = self._request_credentials(self._authorization_key)
                self.save()

            except Exception as e:
                self.error(f'failed to obtain credentials: {e}')
                raise

        headers = headers or {}
        headers['Authorization'] = f'Bearer {self._credentials["access_token"]}'

        self.debug(f'requesting {url}')
        request = Request(url, data=body, headers=headers)
        try:
            response = utils.urlopen(request)

        except HTTPError as e:
            if (
                e.code == 401 and retry_auth
            ):  # unauthorized, access token may have expired
                try:
                    self.debug('credentials have probably expired, refreshing them')
                    self._credentials = self._refresh_credentials(
                        self._credentials['refresh_token']
                    )
                    self.save()

                    # retry the request with refreshed credentials
                    return self._request(url, body, headers, retry_auth=False)

                except Exception:
                    self.error('refreshing credentials failed')
                    raise

            elif str(e).count('not_found'):
                msg = f'folder "{self._location}" not found'
                self.error(msg)
                raise Exception(msg)

            else:
                self.error(f'request failed: {e}')
                raise

        except Exception as e:
            self.error(f'request failed: {e}')
            raise

        return response.read()

    def _request_credentials(self, authorization_key):
        headers = {'Content-Type': 'application/x-www-form-urlencoded'}

        body = {
            'code': authorization_key,
            'client_id': self.CLIENT_ID,
            'client_secret': self.CLIENT_NOT_SO_SECRET,
            'grant_type': 'authorization_code',
        }
        body = urlencode(body).encode()

        request = Request(self.TOKEN_URL, data=body, headers=headers)

        try:
            response = utils.urlopen(request)

        except HTTPError as e:
            error = json.load(e)
            raise Exception(
                error.get('error_description') or error.get('error') or str(e)
            )

        data = json.load(response)

        return {
            'access_token': data['access_token'],
            'refresh_token': data['refresh_token'],
        }

    def _refresh_credentials(self, refresh_token):
        headers = {'Content-Type': 'application/x-www-form-urlencoded'}

        body = {
            'refresh_token': refresh_token,
            'client_id': self.CLIENT_ID,
            'client_secret': self.CLIENT_NOT_SO_SECRET,
            'grant_type': 'refresh_token',
        }
        body = urlencode(body).encode()

        request = Request(self.TOKEN_URL, data=body, headers=headers)

        try:
            response = utils.urlopen(request)

        except HTTPError as e:
            error = json.load(e)
            raise Exception(
                error.get('error_description') or error.get('error') or str(e)
            )

        data = json.load(response)

        return {
            'access_token': data['access_token'],
            'refresh_token': data.get('refresh_token', refresh_token),
        }


class Webdav(UploadService):
    NAME = 'webdav'

    def __init__(self, camera_id):
        self._endpoint_url = None
        self._username = None
        self._password = None
        self._location = None

        UploadService.__init__(self, camera_id)

    def _request(self, url, method, body=None):
        base64string = b64encode(f'{self._username}:{self._password}'.encode()).decode(
            'ascii'
        )
        headers = {'Authorization': f'Basic {base64string}'}
        if body is not None:
            headers.update({'Content-Length': len(body)})
        self.debug(f'request: {method} {url}')
        request = Request(url, data=body, headers=headers)
        request.get_method = lambda: method
        try:
            utils.urlopen(request)
        except HTTPError as e:
            if method == 'MKCOL' and e.code == 405:
                self.debug(
                    'MKCOL failed with code 405, this is normal if the folder exists'
                )
            else:
                raise e

    def _make_dirs(self, path):
        dir_url = self._endpoint_url.rstrip('/') + '/'
        for folder in path.split('/'):
            dir_url = dir_url + folder + '/'
            self._request(dir_url, 'MKCOL')

    def test_access(self):
        try:
            path = self._location.strip('/') + '/' + str(time.time())
            self._make_dirs(path)
            self._request(self._endpoint_url.rstrip('/') + '/' + path, 'DELETE')
            return True
        except Exception as e:
            self.error(str(e), exc_info=True)
            return str(e)

    def upload_data(self, filename, mime_type, data, ctime, camera_name):
        path = self._location.strip('/') + '/' + os.path.dirname(filename)
        filename = os.path.basename(filename)
        self._make_dirs(path)
        self.debug(f'uploading {filename} of {len(data)} bytes')
        self._request(
            self._endpoint_url.rstrip('/') + '/' + path + '/' + filename,
            'PUT',
            bytearray(data),
        )
        self.debug('upload done')

    def dump(self):
        return {
            'endpoint_url': self._endpoint_url,
            'username': self._username,
            'password': self._password,
            'location': self._location,
        }

    def load(self, data):
        if data.get('endpoint_url') is not None:
            self._endpoint_url = data['endpoint_url']
        if data.get('username') is not None:
            self._username = data['username']
        if data.get('password') is not None:
            self._password = data['password']
        if data.get('location'):
            self._location = data['location']


class FTP(UploadService):
    NAME = 'ftp'
    CONN_LIFE_TIME = 60  # don't keep an FTP connection for more than 1 minute

    def __init__(self, camera_id):
        self._server = None
        self._port = None
        self._username = None
        self._password = None
        self._location = None

        self._conn = None
        self._conn_time = 0

        UploadService.__init__(self, camera_id)

    def test_access(self):
        try:
            conn = self._get_conn(create=True)

            path = self._make_dirs(self._location, conn=conn)
            conn.cwd(path)

            d = '%s' % int(time.time())
            self.debug(f'creating test directory {path}/{d}')
            conn.mkd(d)
            conn.rmd(d)

            return True

        except Exception as e:
            self.error(str(e), exc_info=True)

            return str(e)

    def upload_data(self, filename, mime_type, data, ctime, camera_name):
        path = os.path.dirname(filename)
        filename = os.path.basename(filename)

        conn = self._get_conn()
        path = self._make_dirs(self._location + '/' + path, conn=conn)
        conn.cwd(path)

        self.debug(f'uploading {filename} of {len(data)} bytes')
        conn.storbinary(f'STOR {filename}', io.BytesIO(data))

        self.debug('upload done')

    def dump(self):
        return {
            'server': self._server,
            'port': self._port,
            'username': self._username,
            'password': self._password,
            'location': self._location,
        }

    def load(self, data):
        if data.get('server') is not None:
            self._server = data['server']
        if data.get('port') is not None:
            self._port = int(data['port'])
        if data.get('username') is not None:
            self._username = data['username']
        if data.get('password') is not None:
            self._password = data['password']
        if data.get('location'):
            self._location = data['location']

    def _get_conn(self, create=False):
        now = time.time()
        if self._conn is None or now - self._conn_time > self.CONN_LIFE_TIME or create:
            self.debug(
                'creating connection to {}@{}:{}'.format(
                    self._username or 'anonymous', self._server, self._port
                )
            )
            self._conn = ftplib.FTP()
            self._conn.set_pasv(True)
            self._conn.connect(self._server, port=self._port)
            self._conn.login(self._username or 'anonymous', self._password)
            self._conn_time = now

        return self._conn

    def _make_dirs(self, path, conn=None):
        conn = conn or self._get_conn()

        path = path.split('/')
        path = [p for p in path if p]

        self.debug('ensuring path /%s' % '/'.join(path))

        conn.cwd('/')
        for p in path:
            l = conn.nlst()
            if p not in l:
                conn.mkd(p)

            conn.cwd(p)

        return '/' + '/'.join(path)


class SFTP(UploadService):
    NAME = 'sftp'

    def __init__(self, camera_id):
        self._server = None
        self._port = None
        self._username = None
        self._password = None
        self._location = None

        UploadService.__init__(self, camera_id)

    def curl_perform_filetransfer(self, conn):
        curl_url = conn.getinfo(pycurl.EFFECTIVE_URL)

        try:
            conn.perform()

        except pycurl.error:
            curl_error = conn.errstr()
            msg = f'cURL upload failed on {curl_url}: {curl_error}'
            self.error(msg)
            raise

        else:
            self.debug(f'upload done: {curl_url}')

        finally:
            conn.close()

    def test_access(self):
        filename = time.time()
        test_folder = "motioneye_test"
        test_file = f"/{test_folder}/{filename}"

        # list of commands to send after upload.
        rm_operations = [
            f'RM {self._location}/{test_file}',
            f'RMDIR {self._location}/{test_folder}',
        ]

        conn = self._get_conn(test_file)
        conn.setopt(conn.POSTQUOTE, rm_operations)  # Executed after transfer.
        conn.setopt(pycurl.READFUNCTION, io.StringIO().read)

        try:
            self.curl_perform_filetransfer(conn)
            return True

        except Exception as e:
            logging.error(f'sftp connection failed: {e}')

            return str(e)

    def upload_data(self, filename, mime_type, data, ctime, camera_name):
        conn = self._get_conn(filename)
        conn.setopt(pycurl.READFUNCTION, io.BytesIO(data).read)

        self.curl_perform_filetransfer(conn)

    def dump(self):
        return {
            'server': self._server,
            'port': self._port,
            'username': self._username,
            'password': self._password,
            'location': self._location,
        }

    def load(self, data):
        if data.get('server') is not None:
            self._server = data['server']
        if data.get('port') is not None:
            self._port = int(data['port'])
        if data.get('username') is not None:
            self._username = data['username']
        if data.get('password') is not None:
            self._password = data['password']
        if data.get('location'):
            self._location = data['location']

    def _get_conn(self, filename, auth_type='password'):
        sftp_url = 'sftp://{}:{}/{}/{}'.format(
            self._server, self._port, self._location, filename
        )

        self.debug(
            'creating sftp connection to {}@{}:{}'.format(
                self._username, self._server, self._port
            )
        )

        self._conn = pycurl.Curl()
        self._conn.setopt(self._conn.URL, sftp_url)
        self._conn.setopt(pycurl.CONNECTTIMEOUT, 10)
        self._conn.setopt(
            self._conn.FTP_CREATE_MISSING_DIRS, 2
        )  # retry once if MKD fails

        auth_types = {
            'password': self._conn.SSH_AUTH_PASSWORD,
            # 'private_key': self._conn.SSH_PRIVATE_KEYFILE
            # ref: https://curl.haxx.se/libcurl/c/CURLOPT_SSH_PRIVATE_KEYFILE.html
        }

        try:
            self._conn.setopt(self._conn.SSH_AUTH_TYPES, auth_types[auth_type])

        except KeyError:
            self.error(f"invalid SSH auth type: {auth_type}")
            raise

        if auth_type == 'password':
            self._conn.setopt(self._conn.USERNAME, self._username)
            self._conn.setopt(self._conn.PASSWORD, self._password)

        self._conn.setopt(self._conn.UPLOAD, 1)

        return self._conn


class S3(UploadService):
    NAME = 's3'

    def __init__(self, camera_id):
        self._endpoint_url = None
        self._access_key = None
        self._secret_key = None
        self._bucket = None
        UploadService.__init__(self, camera_id)

    def dump(self):
        return {
            'endpoint_url': self._endpoint_url,
            'access_key': self._access_key,
            'secret_key': self._secret_key,
            'bucket': self._bucket,
        }

    def load(self, data):
        if data.get('endpoint_url'):
            self._endpoint_url = data['endpoint_url']
        if data.get('access_key') is not None:
            self._access_key = data['access_key']
        if data.get('secret_key') is not None:
            self._secret_key = data['secret_key']
        if data.get('bucket') is not None:
            self._bucket = data['bucket']

    def upload_file(self, target_dir, filename, camera_name):
        # Create an S3 client
        s3 = boto3.client(
            's3',
            endpoint_url=self._endpoint_url,
            aws_access_key_id=self._access_key,
            aws_secret_access_key=self._secret_key,
        )

        # Uploads the given file using a managed uploader, which will split up
        # large files automatically and upload parts in parallel.
        self.debug(f'uploading file "{filename}" to S3 bucket "{self._bucket}"')
        s3.upload_file(filename, self._bucket, filename[len(target_dir) :])

    def test_access(self):
        try:
            # Create an S3 client
            s3 = boto3.client(
                's3',
                endpoint_url=self._endpoint_url,
                aws_access_key_id=self._access_key,
                aws_secret_access_key=self._secret_key,
            )
            response = s3.list_buckets()
            logging.debug('Existing buckets:')
            for bucket in response['Buckets']:
                logging.debug(f'  {bucket["Name"]}')
            return True
        except Exception as e:
            logging.error(f'S3 connection failed: {e}')
            return str(e)


def get_authorize_url(service_name):
    cls = UploadService.get_service_classes().get(service_name)

    if cls:
        return cls.get_authorize_url()

    else:
        return None


def get(camera_id, service_name):
    global _services

    if _services is None:
        _services = _load()

    camera_id = str(camera_id)

    service = _services.get(camera_id, {}).get(service_name)

    if service is None:
        cls = UploadService.get_service_classes().get(service_name)

        if cls:
            service = cls(camera_id=camera_id)
            _services.setdefault(camera_id, {})[service_name] = service

            logging.debug(
                f'created default upload service "{service_name}" for camera with id "{camera_id}"'
            )

    return service


def test_access(camera_id, service_name, data):
    logging.debug(f'testing access to {service_name}')

    service = get(camera_id, service_name)
    service.load(data)
    if not service:
        return f'unknown upload service {service_name}'

    return service.test_access()


def update(camera_id, service_name, settings):
    service = get(camera_id, service_name)
    service.load(settings)
    service.save()


def upload_media_file(camera_id, camera_name, target_dir, service_name, filename):
    service = get(camera_id, service_name)
    if not service:
        return logging.error(
            f'service "{service_name}" not initialized for camera with id {camera_id}'
        )

    try:
        service.upload_file(target_dir, filename, camera_name)

    except Exception as e:
        logging.error(
            f'failed to upload file "{filename}" with service {service}: {e}',
            exc_info=True,
        )


def _load():
    services = {}

    file_path = os.path.join(settings.CONF_PATH, _STATE_FILE_NAME)

    if os.path.exists(file_path):
        logging.debug(f'loading upload services state from "{file_path}"...')

        try:
            f = open(file_path)

        except Exception as e:
            logging.error(
                f'could not open upload services state file "{file_path}": {e}'
            )

            return services

        try:
            data = json.load(f)

        except Exception as e:
            logging.error(
                f'could not read upload services state from file "{file_path}": {e}'
            )

            return services

        finally:
            f.close()

        for camera_id, d in list(data.items()):
            for name, state in list(d.items()):
                camera_services = services.setdefault(camera_id, {})
                cls = UploadService.get_service_classes().get(name)

                if cls:
                    service = cls(camera_id=camera_id)
                    service.load(state)

                    camera_services[name] = service

                    logging.debug(
                        f'loaded upload service "{name}" for camera with id "{camera_id}"'
                    )

    return services


def _save(services):
    file_path = os.path.join(settings.CONF_PATH, _STATE_FILE_NAME)

    logging.debug(f'saving upload services state to "{file_path}"...')

    data = {}
    for camera_id, camera_services in list(services.items()):
        for name, service in list(camera_services.items()):
            data.setdefault(str(camera_id), {})[name] = service.dump()

    try:
        f = open(file_path, 'w')

    except Exception as e:
        logging.error(f'could not open upload services state file "{file_path}": {e}')

        return

    try:
        json.dump(data, f, sort_keys=True, indent=4)

    except Exception as e:
        logging.error(
            f'could not save upload services state to file "{file_path}": {e}'
        )

    finally:
        f.close()


def clean_cloud(local_dir, data, info):
    camera_id = info['camera_id']
    service_name = info['service_name']
    cloud_dir_user = info['cloud_dir']
    cloud_dir = [p.strip() for p in cloud_dir_user.split('/') if p.strip()][0]

    logging.debug(f'clean_cloud({camera_id}, {service_name}, {local_dir}, {cloud_dir})')

    if service_name and local_dir and cloud_dir:
        local_folders = get_local_folders(local_dir)
        service = get(camera_id, service_name)
        service.load(data)
        service.clean_cloud(cloud_dir, local_folders)


def exist_in_local(folder, local_folders):
    if not local_folders:
        local_folders = []

    if not folder:
        return False

    return folder in local_folders


def get_local_folders(dirpath):
    folders = next(os.walk(dirpath))[1]
    return folders



# Copyright (c) 2013 Calin Crisan
# This file is part of motionEye.
#
# motionEye is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import errno
import logging
import re
import socket
import time
from typing import Any, Tuple

from tornado.concurrent import Future
from tornado.ioloop import IOLoop
from tornado.iostream import IOStream

from motioneye import config, motionctl, settings, utils


class MjpgClient(IOStream):
    _FPS_LEN = 10

    clients = {}  # dictionary of clients indexed by camera id
    _last_erroneous_close_time = (
        0  # helps detecting erroneous connections and restart motion
    )

    def __init__(self, camera_id, port, username, password, auth_mode):
        self._camera_id = camera_id
        self._port = port
        self._username = username or ''
        self._password = password or ''
        self._auth_mode = auth_mode
        self._auth_digest_state = {}

        self._last_access = 0
        self._last_jpg = None
        self._last_jpg_times = []

        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
        IOStream.__init__(self, s)

        self.set_close_callback(self.on_close)

    def do_connect(self) -> 'Future[MjpgClient]':
        f = self.connect(('localhost', self._port))
        f.add_done_callback(self._on_connect)
        return f

    def get_port(self):
        return self._port

    def on_close(self):
        logging.debug(
            f'connection closed for mjpg client for camera {self._camera_id} on port {self._port}'
        )

        if MjpgClient.clients.pop(self._camera_id, None):
            logging.debug(
                f'mjpg client for camera {self._camera_id} on port {self._port} removed'
            )

        if getattr(self, 'error', None) and self.error.errno != errno.ECONNREFUSED:
            now = time.time()
            if (
                now - MjpgClient._last_erroneous_close_time
                < settings.MJPG_CLIENT_TIMEOUT
            ):
                msg = f'connection problem detected for mjpg client for camera {self._camera_id} on port {self._port}'

                logging.error(msg)

                if settings.MOTION_RESTART_ON_ERRORS:
                    motionctl.stop(
                        invalidate=True
                    )  # this will close all the mjpg clients
                    motionctl.start(deferred=True)

            MjpgClient._last_erroneous_close_time = now

    def get_last_jpg(self):
        self._last_access = time.time()
        return self._last_jpg

    def get_last_access(self):
        return self._last_access

    def get_last_jpg_time(self):
        if not self._last_jpg_times:
            self._last_jpg_times.append(time.time())

        return self._last_jpg_times[-1]

    def get_fps(self):
        if len(self._last_jpg_times) < self._FPS_LEN:
            return 0  # not enough "samples"

        if time.time() - self._last_jpg_times[-1] > 1:
            return 0  # everything below 1 fps is considered 0

        return (len(self._last_jpg_times) - 1) / (
            self._last_jpg_times[-1] - self._last_jpg_times[0]
        )

    def _check_error(self) -> bool:
        if self.socket is None:
            logging.warning(
                f'mjpg client connection for camera {self._camera_id} on port {self._port} is closed'
            )

            self.close()

            return True

        error = getattr(self, 'error', None)
        if (error is None) or (
            getattr(error, 'errno', None) == 0
        ):  # error could also be ESUCCESS for some reason
            return False

        self._error(error)

        return True

    def _error(self, error) -> None:
        logging.error(
            f'mjpg client for camera {self._camera_id} on port {self._port} error: {str(error)}',
            exc_info=True,
        )

        try:
            self.close()

        except Exception:
            pass

    def _get_future_result(self, future: Future) -> Tuple[bool, Any]:
        try:
            if self.closed():
                future.cancel()
                logging.debug('_on_connect: Stream is closed. Future cancelled.')
                return False, None

            return True, future.result()
        except Exception as e:
            self._error(e)
            return False, None

    def _on_connect(self, future: Future) -> None:
        result, _ = self._get_future_result(future)
        if not result:
            return

        logging.debug(
            f'mjpg client for camera {self._camera_id} connected on port {self._port}'
        )

        if self._auth_mode == 'basic':
            logging.debug('mjpg client using basic authentication')
            auth_header = utils.build_basic_header(self._username, self._password)
            self.write(
                f'GET / HTTP/1.0\r\nAuthorization: {auth_header}\r\nConnection: close\r\n\r\n'.encode()
            )

        elif (
            self._auth_mode == 'digest'
        ):  # in digest auth mode, the header is built upon receiving 401
            logging.debug('digest authentication _on_connect')
            self.write(b'GET / HTTP/1.0\r\n\r\n')

        else:  # no authentication
            logging.debug('no authentication _on_connect')
            self.write(b'GET / HTTP/1.0\r\nConnection: close\r\n\r\n')

        self._seek_http()

    def _seek_http(self, future: Future = False) -> None:
        result, _ = self._get_future_result(future) if future else (True, False)

        if not result or self._check_error():
            return

        future = utils.cast_future(self.read_until_regex(br'HTTP/1.\d \d+ '))
        future.add_done_callback(self._on_http)

    def _on_http(self, future: Future) -> None:
        result, data = self._get_future_result(future)
        if not result:
            return

        if data.endswith(b'401 '):
            self._seek_www_authenticate()

        else:  # no authorization required, skip to content length
            self._seek_content_length()

    def _seek_www_authenticate(self) -> None:
        future = utils.cast_future(self.read_until(b'WWW-Authenticate:'))
        future.add_done_callback(self._on_before_www_authenticate)

    def _on_before_www_authenticate(self, future: Future) -> None:
        result, _ = self._get_future_result(future)
        if not result or self._check_error():
            return

        r_future = utils.cast_future(self.read_until(b'\r\n'))
        r_future.add_done_callback(self._on_www_authenticate)

    def _on_www_authenticate(self, future: Future) -> None:
        result, data = self._get_future_result(future)
        if not result or self._check_error():
            return

        data = data.strip()

        m = re.match(br'Basic\s*realm="([a-zA-Z0-9\-\s]+)"', data)
        if m:
            logging.debug('mjpg client using basic authentication')

            auth_header = utils.build_basic_header(self._username, self._password)
            w_data = f'GET / HTTP/1.0\r\nAuthorization: {auth_header}\r\nConnection: close\r\n\r\n'.encode()
            w_future = utils.cast_future(self.write(w_data))
            w_future.add_done_callback(self._seek_http)

            return

        data = data.decode()
        if data.startswith('Digest'):
            logging.debug('mjpg client using digest authentication')

            parts = data[7:].split(',')
            parts_dict = dict(p.split('=', 1) for p in parts)
            parts_dict = {p[0]: p[1].strip('"') for p in list(parts_dict.items())}

            self._auth_digest_state = parts_dict

            auth_header = utils.build_digest_header(
                'GET', '/', self._username, self._password, self._auth_digest_state
            )
            w_data = f'GET / HTTP/1.0\r\nAuthorization: {auth_header}\r\nConnection: close\r\n\r\n'.encode()
            w_future = utils.cast_future(self.write(w_data))
            w_future.add_done_callback(self._seek_http)

            return

        logging.error(f'mjpg client unknown authentication header: {data}')
        self._seek_content_length()

    def _seek_content_length(self):
        if self._check_error():
            return

        r_future = utils.cast_future(self.read_until(b'Content-Length:'))
        r_future.add_done_callback(self._on_before_content_length)

    def _on_before_content_length(self, future: Future):
        result, _ = self._get_future_result(future)
        if not result or self._check_error():
            return

        r_future = utils.cast_future(self.read_until(b'\r\n\r\n'))
        r_future.add_done_callback(self._on_content_length)

    def _on_content_length(self, future: Future):
        result, data = self._get_future_result(future)
        if not result or self._check_error():
            return

        matches = re.findall(rb'(\d+)', data)
        if not matches:
            self._error(f'could not find content length in mjpg header line "{data}"')

            return

        length = int(matches[0])

        r_future = utils.cast_future(self.read_bytes(length))
        r_future.add_done_callback(self._on_jpg)

    def _on_jpg(self, future: Future):
        result, data = self._get_future_result(future)
        if not result:
            return

        self._last_jpg = data
        self._last_jpg_times.append(time.time())
        while len(self._last_jpg_times) > self._FPS_LEN:
            self._last_jpg_times.pop(0)

        self._seek_content_length()


def start():
    # schedule the garbage collector
    io_loop = IOLoop.current()
    io_loop.add_timeout(
        datetime.timedelta(seconds=settings.MJPG_CLIENT_TIMEOUT), _garbage_collector
    )


def get_jpg(camera_id):
    if camera_id not in MjpgClient.clients:
        # mjpg client not started yet for this camera

        logging.debug(f'creating mjpg client for camera {camera_id}')

        camera_config = config.get_camera(camera_id)
        if not camera_config['@enabled'] or not utils.is_local_motion_camera(
            camera_config
        ):
            logging.error(
                f'could not start mjpg client for camera id {camera_id}: not enabled or not local'
            )

            return None

        port = camera_config['stream_port']
        username, password = None, None
        auth_mode = None
        if camera_config.get('stream_auth_method') > 0:
            username, password = camera_config.get('stream_authentication', ':').split(
                ':'
            )
            auth_mode = (
                'digest' if camera_config.get('stream_auth_method') > 1 else 'basic'
            )

        client = MjpgClient(camera_id, port, username, password, auth_mode)
        client.do_connect()

        MjpgClient.clients[camera_id] = client

    client = MjpgClient.clients[camera_id]

    return client.get_last_jpg()


def get_fps(camera_id):
    client = MjpgClient.clients.get(camera_id)
    if client is None:
        return 0

    return client.get_fps()


def close_all(invalidate=False):
    for client in list(MjpgClient.clients.values()):
        client.close()

    if invalidate:
        MjpgClient.clients = {}
        MjpgClient._last_erroneous_close_time = 0


def _garbage_collector():
    io_loop = IOLoop.current()
    io_loop.add_timeout(
        datetime.timedelta(seconds=settings.MJPG_CLIENT_TIMEOUT), _garbage_collector
    )

    now = time.time()
    for camera_id, client in list(MjpgClient.clients.items()):
        logging.debug(f'_garbage_collector checking camera. id: {camera_id}')
        port = client.get_port()

        if client.closed():
            continue

        # check for jpeg frame timeout
        last_jpg_time = client.get_last_jpg_time()
        delta = now - last_jpg_time
        if delta > settings.MJPG_CLIENT_TIMEOUT:
            logging.error(
                f'mjpg client timed out receiving data for camera {camera_id} on port {port}'
            )

            if settings.MOTION_RESTART_ON_ERRORS:
                motionctl.stop(invalidate=True)  # this will close all the mjpg clients
                motionctl.start(deferred=True)

            break

        # check for last access timeout
        delta = now - client.get_last_access()
        if (
            settings.MJPG_CLIENT_IDLE_TIMEOUT
            and delta > settings.MJPG_CLIENT_IDLE_TIMEOUT
        ):
            msg = (
                'mjpg client for camera %(camera_id)s on port %(port)s has been idle '
                'for %(timeout)s seconds, removing it'
                % {
                    'camera_id': camera_id,
                    'port': port,
                    'timeout': settings.MJPG_CLIENT_IDLE_TIMEOUT,
                }
            )

            logging.debug(msg)

            client.close()

            continue



# Copyright (c) 2020 Vlsarro
# Copyright (c) 2013 Calin Crisan
# This file is part of motionEye.
#
# motionEye is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import logging

from tornado.ioloop import IOLoop

from motioneye import config, motionctl, utils


def _start_check_ws() -> None:
    IOLoop.current().spawn_callback(_check_ws)


def start() -> None:
    io_loop = IOLoop.current()
    io_loop.add_timeout(datetime.timedelta(seconds=1), _start_check_ws)


def _during_working_schedule(now, working_schedule) -> bool:
    parts = working_schedule.split('|')
    if len(parts) < 7:
        return False  # invalid ws

    ws_day = parts[now.weekday()]
    parts = ws_day.split('-')
    if len(parts) != 2:
        return False  # invalid ws

    _from, to = parts
    if not _from or not to:
        return False  # ws disabled for this day

    _from = _from.split(':')
    to = to.split(':')
    if len(_from) != 2 or len(to) != 2:
        return False  # invalid ws

    try:
        from_h = int(_from[0])
        from_m = int(_from[1])
        to_h = int(to[0])
        to_m = int(to[1])

    except ValueError:
        return False  # invalid ws

    if now.hour < from_h or now.hour > to_h:
        return False

    if now.hour == from_h and now.minute < from_m:
        return False

    if now.hour == to_h and now.minute > to_m:
        return False

    return True


async def _switch_motion_detection_status(
    camera_id,
    must_be_enabled,
    working_schedule_type,
    motion_detection_resp: utils.GetMotionDetectionResult,
) -> None:
    if motion_detection_resp.error:  # could not detect current status
        return logging.warning(
            'skipping motion detection status update for camera with id {id}: {error}'.format(
                id=camera_id, error=motion_detection_resp.error
            )
        )

    if motion_detection_resp.enabled and not must_be_enabled:
        logging.debug(
            'must disable motion detection for camera with id {id} ({what} working schedule)'.format(
                id=camera_id, what=working_schedule_type
            )
        )

        await motionctl.set_motion_detection(camera_id, False)

    elif not motion_detection_resp.enabled and must_be_enabled:
        logging.debug(
            'must enable motion detection for camera with id {id} ({what} working schedule)'.format(
                id=camera_id, what=working_schedule_type
            )
        )

        await motionctl.set_motion_detection(camera_id, True)


async def _check_ws() -> None:
    # schedule the next call
    io_loop = IOLoop.current()
    io_loop.add_timeout(datetime.timedelta(seconds=10), _start_check_ws)

    if not motionctl.running():
        return

    now = datetime.datetime.now()
    for camera_id in config.get_camera_ids():
        camera_config = config.get_camera(camera_id)
        if not utils.is_local_motion_camera(camera_config):
            continue

        working_schedule = camera_config.get('@working_schedule')
        motion_detection = camera_config.get('@motion_detection')
        working_schedule_type = camera_config.get('@working_schedule_type') or 'outside'

        if (
            not working_schedule
        ):  # working schedule disabled, motion detection left untouched
            continue

        if not motion_detection:  # motion detection explicitly disabled
            continue

        now_during = _during_working_schedule(now, working_schedule)
        must_be_enabled = (now_during and working_schedule_type == 'during') or (
            not now_during and working_schedule_type == 'outside'
        )

        motion_detection_resp = await motionctl.get_motion_detection(camera_id)
        await _switch_motion_detection_status(
            camera_id, must_be_enabled, working_schedule_type, motion_detection_resp
        )



# Copyright (c) 2020 Vlsarro
# Copyright (c) 2013 Calin Crisan
# This file is part of motionEye.
#
# motionEye is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from motioneye import config, motionctl, settings, update, utils
from motioneye.handlers.base import BaseHandler

__all__ = ('MainHandler',)


class MainHandler(BaseHandler):
    def get(self):
        # additional config
        main_sections = config.get_additional_structure(camera=False, separators=True)[
            0
        ]
        camera_sections = config.get_additional_structure(camera=True, separators=True)[
            0
        ]

        motion_info = motionctl.find_motion()
        os_version = update.get_os_version()

        self.render(
            'main.html',
            frame=False,
            motion_version=motion_info[1] if motion_info else '(none)',
            os_version=' '.join(os_version),
            enable_update=settings.ENABLE_UPDATE,
            enable_reboot=settings.ENABLE_REBOOT,
            add_remove_cameras=settings.ADD_REMOVE_CAMERAS,
            main_sections=main_sections,
            camera_sections=camera_sections,
            hostname=settings.SERVER_NAME,
            title=self.get_argument('title', None),
            admin_username=config.get_main().get('@admin_username'),
            has_h264_omx_support=motionctl.has_h264_omx_support(),
            has_h264_v4l2m2m_support=motionctl.has_h264_v4l2m2m_support(),
            has_h264_nvenc_support=motionctl.has_h264_nvenc_support(),
            has_h264_nvmpi_support=motionctl.has_h264_nvmpi_support(),
            has_hevc_nvenc_support=motionctl.has_hevc_nvenc_support(),
            has_hevc_nvmpi_support=motionctl.has_hevc_nvmpi_support(),
            has_h264_qsv_support=motionctl.has_h264_qsv_support(),
            has_hevc_qsv_support=motionctl.has_hevc_qsv_support(),
            has_motion=bool(motionctl.find_motion()[0]),
            mask_width=utils.MASK_WIDTH,
        )



# Copyright (c) 2020 Vlsarro
# Copyright (c) 2013 Calin Crisan
# This file is part of motionEye.
#
# motionEye is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import tempfile
import time

from tornado.web import HTTPError, StaticFileHandler

from motioneye import config, mediafiles, remote, utils
from motioneye.handlers.base import BaseHandler

__all__ = ('MoviePlaybackHandler', 'MovieDownloadHandler')


# support fetching movies with authentication
class MoviePlaybackHandler(StaticFileHandler, BaseHandler):
    tmpdir = tempfile.gettempdir() + '/MotionEye'
    if not os.path.exists(tmpdir):
        os.mkdir(tmpdir)

    @BaseHandler.auth()
    async def get(self, camera_id, filename=None, include_body=True):
        logging.debug(
            'downloading movie {filename} of camera {id}'.format(
                filename=filename, id=camera_id
            )
        )

        self.pretty_filename = os.path.basename(filename)

        if camera_id is not None:
            camera_id = int(camera_id)
            if camera_id not in config.get_camera_ids():
                raise HTTPError(404, 'no such camera')

        camera_config = config.get_camera(camera_id)

        if utils.is_local_motion_camera(camera_config):
            filename = mediafiles.get_media_path(camera_config, filename, 'movie')
            self.pretty_filename = (
                camera_config['camera_name'] + '_' + self.pretty_filename
            )
            await StaticFileHandler.get(self, filename, include_body=include_body)
            return

        elif utils.is_remote_camera(camera_config):
            # we will cache the movie since it takes a while to fetch from the remote camera
            # and we may be going to play it back in the browser, which will fetch the video in chunks
            tmpfile = self.tmpdir + '/' + self.pretty_filename
            if os.path.isfile(tmpfile):
                # have a cached copy, update the timestamp so it's not flushed
                import time

                mtime = os.stat(tmpfile).st_mtime
                os.utime(tmpfile, (time.time(), mtime))
                await StaticFileHandler.get(self, tmpfile, include_body=include_body)
                return

            resp = await remote.get_media_content(
                camera_config, filename, media_type='movie'
            )
            if resp.error:
                return self.finish_json(
                    {
                        'error': 'Failed to download movie from {url}: {msg}.'.format(
                            url=remote.pretty_camera_url(camera_config), msg=resp.error
                        )
                    }
                )

            # check if the file has been created by another request while we were fetching the movie
            if not os.path.isfile(tmpfile):
                tmp = open(tmpfile, 'wb')
                tmp.write(resp.result)
                tmp.close()

            await StaticFileHandler.get(self, tmpfile, include_body=include_body)
            return

        else:  # assuming simple mjpeg camera
            raise HTTPError(400, 'unknown operation')

    def on_finish(self):
        # delete any cached file older than an hour
        stale_time = time.time() - (60 * 60)
        try:
            for f in os.listdir(self.tmpdir):
                f = os.path.join(self.tmpdir, f)
                if os.path.isfile(f) and os.stat(f).st_atime <= stale_time:
                    os.remove(f)
        except:
            logging.error('could not delete temp file', exc_info=True)
            pass

    def get_absolute_path(self, root, path):
        return path

    def validate_absolute_path(self, root, absolute_path):
        return absolute_path


class MovieDownloadHandler(MoviePlaybackHandler):
    def set_extra_headers(self, filename):
        if self.get_status() in (200, 304):
            self.set_header(
                'Content-Disposition',
                'attachment; filename=' + self.pretty_filename + ';',
            )
            self.set_header('Expires', '0')



#!/usr/bin/env python3

# Copyright (c) 2013 Calin Crisan
# This file is part of motionEye.
#
# motionEye is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import gettext
import locale
import logging
import os.path
import pipes
import sys

import babel

from motioneye import config, settings

# make sure motioneye is on python path
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))


# ŝarĝante tradukojn
def load_l10n():
    locale.setlocale(locale.LC_ALL, '')
    pathname = os.path.dirname(__file__)
    main_config = config.get_main()
    lang = main_config.get('@lang')

    # get list of all available languages
    messagefiles = gettext.find(
        'motioneye',
        pathname + '/locale',
        languages=babel.Locale('eo').languages.keys(),
        all=True,
    )
    languages = [path.split('/')[-3] for path in messagefiles]
    languages.append('eo')
    languages.sort()
    settings.langlist = list(
        zip(languages, [babel.Locale.parse(lang).display_name for lang in languages])
    )

    # install lang
    # verify that lang exists
    if lang == 'eo':
        settings.lingvo = 'eo'
    else:
        file = gettext.find('motioneye', pathname + '/locale', languages=[lang])
        if file:
            settings.lingvo = lang
        else:
            # if lang is not defined, try english
            file = gettext.find('motioneye', pathname + '/locale', languages=['en'])
            if file:
                settings.lingvo = 'en'
            else:
                # if english is not defined, use esperanto (should not happen)
                settings.lingvo = 'eo'
    # install lang
    if settings.lingvo != 'eo':
        settings.traduction = gettext.translation(
            'motioneye', pathname + '/locale', languages=[settings.lingvo]
        )
        settings.traduction.install()
    else:
        settings.traduction = gettext
        gettext.install('motioneye')


def find_command(command):
    if command == 'relayevent':
        relayevent_sh = os.path.join(os.path.dirname(__file__), 'scripts/relayevent.sh')

        cmd = relayevent_sh + ' "%s"' % (settings.config_file or '')

    else:
        cmd = __file__
        cmd = sys.executable + ' ' + cmd
        cmd = cmd.replace('-b', '')  # remove server-specific options
        cmd += ' %s ' % command
        cmd += ' '.join([pipes.quote(arg) for arg in sys.argv[2:] if arg not in ['-b']])

    return cmd


def load_settings():
    # parse common command line arguments

    config_file = None
    debug = False

    for i in range(1, len(sys.argv)):
        arg = sys.argv[i]
        next_arg = i < len(sys.argv) - 1 and sys.argv[i + 1]
        if arg == '-c':
            config_file = next_arg

        elif arg == '-d':
            debug = True

    conf_path_given = [False]
    run_path_given = [False]
    log_path_given = [False]
    media_path_given = [False]

    # parse the config file, if given

    def parse_conf_line(line):
        line = line.strip()
        if not line or line.startswith('#'):
            return

        parts = line.split(' ', 1)
        if len(parts) != 2:
            raise Exception('invalid configuration line: %s' % line)

        name, value = parts
        upper_name = name.upper().replace('-', '_')

        if hasattr(settings, upper_name):
            curr_value = getattr(settings, upper_name)

            if upper_name == 'LOG_LEVEL':
                if value == 'quiet':
                    value = 100

                else:
                    value = getattr(logging, value.upper(), logging.DEBUG)

            elif value.lower() == 'true':
                value = True

            elif value.lower() == 'false':
                value = False

            elif isinstance(curr_value, int):
                value = int(value)

            elif isinstance(curr_value, float):
                value = float(value)

            if upper_name == 'CONF_PATH':
                conf_path_given[0] = True

            elif upper_name == 'RUN_PATH':
                run_path_given[0] = True

            elif upper_name == 'LOG_PATH':
                log_path_given[0] = True

            elif upper_name == 'MEDIA_PATH':
                media_path_given[0] = True

            setattr(settings, upper_name, value)

        else:
            logging.warning('unknown configuration option: %s' % name)

    if config_file:
        try:
            with open(config_file) as f:
                for line in f:
                    parse_conf_line(line)

        except Exception as e:
            logging.fatal(f'failed to read settings from "{config_file}": {e}')
            sys.exit(-1)

        # use the config file directory as base dir
        # if not specified otherwise in the config file
        base_dir = os.path.dirname(config_file)
        settings.config_file = config_file

        if not conf_path_given[0]:
            settings.CONF_PATH = base_dir

        if not run_path_given[0]:
            settings.RUN_PATH = base_dir

        if not log_path_given[0]:
            settings.LOG_PATH = base_dir

        if not media_path_given[0]:
            settings.MEDIA_PATH = base_dir

    else:
        logging.info('no configuration file given, using built-in defaults')

    if debug:
        settings.LOG_LEVEL = logging.DEBUG


def configure_logging(cmd, log_to_file=False):
    sys.stderr.write(f'configure_logging cmd {cmd}: {log_to_file}\n')
    if log_to_file or cmd != 'motioneye':
        fmt = f'%(asctime)s: [{cmd}] %(levelname)8s: %(message)s'

    else:
        fmt = f'%(levelname)8s: %(message)s'

    for h in logging.getLogger().handlers:
        logging.getLogger().removeHandler(h)

    try:
        if log_to_file:
            log_file = os.path.join(settings.LOG_PATH, 'motioneye.log')

        else:
            log_file = None

        sys.stderr.write('configure logging to file: %s\n' % log_file)
        logging.basicConfig(
            filename=log_file,
            level=settings.LOG_LEVEL,
            format=fmt,
            datefmt='%Y-%m-%d %H:%M:%S',
        )

    except Exception as e:
        sys.stderr.write('failed to configure logging: %s\n' % e)
        sys.exit(-1)

    logging.getLogger('tornado').setLevel(logging.WARN)
    logging.getLogger('oauth2client').setLevel(logging.WARN)


def configure_tornado():
    from tornado.httpclient import AsyncHTTPClient

    AsyncHTTPClient.configure(
        'tornado.curl_httpclient.CurlAsyncHTTPClient', max_clients=16
    )


def make_arg_parser(command=None):
    if command:
        usage = description = epilog = None

    else:
        usage = '%(prog)s [command] [-c CONFIG_FILE] [-d] [-h] [-l] [-v] [command options...]\n\n'

        description = 'available commands:\n'
        description += '  startserver\n'
        description += '  stopserver\n'
        description += '  sendmail\n'
        description += '  sendtelegram\n'
        description += '  webhook\n'
        description += '  shell\n\n'

        epilog = 'type "%(prog)s [command] -h" for help on a specific command\n\n'

    parser = argparse.ArgumentParser(
        prog='meyectl%s' % ((' ' + command) if command else ''),
        usage=usage,
        description=description,
        epilog=epilog,
        add_help=False,
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument(
        '-c',
        help='use a config file instead of built-in defaults',
        type=str,
        dest='config_file',
    )
    parser.add_argument(
        '-d',
        help='enable debugging, overriding log level from config file',
        action='store_true',
        dest='debug',
    )
    parser.add_argument(
        '-h', help='print this help and exit', action='help', default=argparse.SUPPRESS
    )
    parser.add_argument(
        '-l',
        help='log to file instead of standard error',
        action='store_true',
        dest='log_to_file',
    )
    parser.add_argument(
        '-v',
        help='print program version and exit',
        action='version',
        default=argparse.SUPPRESS,
    )

    return parser


def print_usage_and_exit(code):
    parser = make_arg_parser()
    parser.print_help(sys.stderr)

    sys.exit(code)


def print_version_and_exit():
    import motioneye

    sys.stderr.write('motionEye %s\n' % motioneye.VERSION)
    sys.exit()


def main():
    for a in sys.argv:
        if a == '-v':
            print_version_and_exit()

    if len(sys.argv) < 2 or sys.argv[1] == '-h':
        print_usage_and_exit(0)

    load_settings()

    command = sys.argv[1]
    arg_parser = make_arg_parser(command)

    load_l10n()

    if command in ('startserver', 'stopserver'):
        from motioneye import server

        server.main(arg_parser, sys.argv[2:], command[:-6])

    elif command == 'sendmail':
        from motioneye import sendmail

        sendmail.main(arg_parser, sys.argv[2:])
    elif command == 'sendtelegram':
        from motioneye import sendtelegram

        sendtelegram.main(arg_parser, sys.argv[2:])
    elif command == 'webhook':
        from motioneye import webhook

        webhook.main(arg_parser, sys.argv[2:])

    elif command == 'shell':
        from motioneye import shell

        shell.main(arg_parser, sys.argv[2:])

    else:
        sys.stderr.write('unknown command "%s"\n\n' % command)
        print_usage_and_exit(-1)


if __name__ == '__main__':
    main()
